#include <config.h>

#include <stdlib.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <packet.h>
#include <cleanup.h>
#include <capdev.h>
#include <capture.h>
#include <decode.h>
#include <preproc.h>
#include <plugin.h>

#ifndef lib_pcap_h

#define pcap_file_header	tcpd_file_header
#define pcap_pkthdr		tcpd_pkthdr

struct tcpd_file_header {
	unsigned int		magic;
	unsigned short		version_major;
	unsigned short		version_minor;
	int			thiszone;
	unsigned int		sigfigs;
	unsigned int		snaplen;
	unsigned int		linktype;
};

struct tcpd_pkthdr {
	struct timeval		ts;
	unsigned int		caplen;
	unsigned int		len;

	/* Nothing beyond this point should
	 * be trusted */
	unsigned int		redhat1;
	unsigned int		redhat2;
};
#endif /* lib_pcap_h */

proc_decode_subproto decode_subproto;
proc_preproc_dispatch preproc_dispatch;
proc_serial_number serial_number;

/* File magics */
#define MAGIC_REDHAT	0xa1b2cd34
#define MAGIC_LBL	0xa1b2c3d4

/* This is our own private data */
#define FIXUP_NONE	0
#define FIXUP_REDHAT	-8
#define FIXUP_LBL	8
struct tcpd_priv {
	struct packet	pkt;
	void		*end;
	void		*cur;
	unsigned int	phsiz;
	struct proto	*protocol;
	void		*map;
	unsigned int	map_size;
	int		fd;
	size_t		snaplen;
};

/* Initialise a capture process, we open the file and then
 * return our opaque private data structure to firestorm */
void *tcpd_init(char *interface)
{
	struct pcap_file_header *fh;
	struct stat st;
	struct tcpd_priv *p=malloc(sizeof(struct tcpd_priv));

	if ( !p ) return NULL;

	/* Interface for us is just the filename  */
	if ( (p->fd=open(interface, O_RDONLY)) < 0 ) {
		printf("tcpdump: %s: open(): %s\n",
			interface, get_err());
		free(p);
		return NULL;	
	}

	/* get the size of the file so we can map it in  */
	if ( fstat(p->fd, &st) ) {
		printf("tcpdump: %s: fstat(): %s\n",
			interface, get_err());
		close(p->fd);
		free(p);
		return NULL;
	}

	/* Check the file is actually even big enough */
	if ( st.st_size < sizeof(struct pcap_file_header) ) {
		printf("tcpdump: %s: Not a valid libpcap file\n",
			interface);
		close(p->fd);
		free(p);
		return NULL;
	}

	/* Memory map that sucker */
	p->map_size=st.st_size;
	if ( (p->map=mmap(NULL, p->map_size, PROT_READ,
		MAP_SHARED, p->fd, 0))==MAP_FAILED) {
		printf("tcpdump: %s: mmap(): %s\n",
			interface, get_err());
		close(p->fd);
		free(p);
		return NULL;
	}
	p->end=p->map+st.st_size;

	fh=(struct pcap_file_header *)p->map;

	/* Check what format the file is */
	switch (fh->magic) {
		case MAGIC_REDHAT: /* Redhat extended shit */
			p->phsiz=24;
			break;
		case MAGIC_LBL: /* LBL Original */
			p->phsiz=16;
			break;
		default: /* WTF! */
			printf("tcpdump: %s: Bad voodoo magic (0x%x)\n",
				interface, fh->magic);
			munmap(p->map, p->map_size);
			close(p->fd);
			free(p);
			return NULL;
	}

	/* Make sure we can decode this link type, not much point
	 * carrying on if we can't decode anything ;)  */
	if ( !(p->protocol=decode_subproto("__pcap_dlt", fh->linktype)) ) {
		printf("tcpdump: %s: Unknown link type (0x%x)\n",
			interface, fh->linktype);
		munmap(p->map, p->map_size);
		close(p->fd);
		free(p);
		return NULL;
	}

	/* Initialise final few fields from header */
	p->snaplen=fh->snaplen;
	p->cur=p->map;
	p->cur+=sizeof(struct pcap_file_header);
	
	return p;
}

/* Stop the capture and cleanup our process */
void tcpd_end(void *priv)
{
	struct tcpd_priv *p=priv;

	if ( !p ) return;

	if ( p->map ) munmap(p->map, p->map_size);
	if ( p->fd>=0 ) close(p->fd);

	free(p);
}

void tcpd_go(void *priv, struct cap_s *c)
{
	struct pcap_pkthdr *h;
	struct tcpd_priv *p=priv;

	if ( !p ) return;

	p->pkt.capture=c;
	p->pkt.flags=FP_PROMISC;

	for(;c->state != CAP_STATE_STOP;) {
		serial_number(&p->pkt.serial);

		/* Make sure a packet header is present */
		if ( p->cur+p->phsiz > p->end ) {
			return;
		}

		/* Advance the p->cur to be the start of the
		 * actual packet data */
		h=(struct pcap_pkthdr *)p->cur;
		p->cur+=p->phsiz;

		/* Check the packet is present */
		if ( (p->cur+h->caplen) > (p->end) ) {
			return;
		}

		/* Fill in the struct packet stuff */
		p->pkt.time.tv_sec=h->ts.tv_sec;
		p->pkt.time.tv_usec=h->ts.tv_usec;
		p->pkt.len=h->len;
		p->pkt.caplen=h->caplen;

		p->pkt.base=p->cur;
		p->pkt.end=p->cur+h->caplen;

		/* advance the file pointer */
		p->cur+=h->caplen;

		/* Dispatch to the firestorm decode engine */
		p->pkt.layer[0].proto=p->protocol;
		p->pkt.layer[0].h.raw=p->pkt.base;
		p->pkt.layer[0].flags=0;
		p->pkt.layer[0].session=NULL;
		p->pkt.llen=1;
		p->pkt.rlen=0;
		p->pkt.present=0;
		if ( p->protocol->decode(&p->pkt) ) {
			preproc_dispatch(&p->pkt);
		}
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	int ok=0;
	proc_capdev_add capdev_add;

	/* validate input */
	plugin_check(in, out);

	/* tell firestorm who we are */
	PLUGIN_ID("capture.tcpdump", "Reads packets in from tcpdump files");
	PLUGIN_VERSION(2, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");

	/* import the function for adding capdevs */
	if ( !(capdev_add=in->import("capdev.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	/* We need this to resolve linktypes */
	if ( !(decode_subproto=in->import("decode.subproto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	/* and this one is to dispatch packets to the
	 * rest of firestorm for processing */
	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(serial_number=in->import("serial_number")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	/* This plugin only provides one service, so bail
	 * if we can't register it */
	ok+=capdev_add("tcpdump", tcpd_init, tcpd_end, tcpd_go);
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL ;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
