/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
*    This file represents the capture subsystem. All capture
*    processes are managed from here. The actual capturing is
*    done in the capdev plugins.
*/
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <cleanup.h>
#include <packet.h>
#include <capture.h>
#include <capdev.h>

struct cap_s *captures=NULL;

void capture_cleanup(int state, void *priv)
{
	struct cap_s *dev;
	struct cap_s *bar;

	printf("capture: stopping all capture processes\n");

	for(dev=captures; dev; ) {
		bar=dev;
		dev=dev->next;
		
		if ( bar->priv ) {
			bar->capdev->end(bar->priv);
		}
		if ( bar->interface ) free(bar->interface);
		free(bar);
	}
}

void capture_stop_all(void)
{
	struct cap_s *cap;

	for(cap=captures; cap; cap=cap->next) {
		cap->state=CAP_STATE_STOP;
	}
}

void capture_init(void)
{
	cleanup_add(capture_cleanup, NULL);
}

int capture_add(const char *type, const char *interface)
{
	struct cap_s *cap;
	
	if ( !type || !interface ) return 0;
	if ( !(cap=calloc(1, sizeof(struct cap_s))) ) {
		cperror("calloc");
	}

	if ( !(cap->capdev=capdev_find(type)) ) {
		return 0;
	}
	
	if ( !(cap->interface=strdup(interface)) ) {
		cperror("strdup");
	}
	
	cap->next=captures;
	cap->state=CAP_STATE_NULL;
	captures=cap;
	
	return 1;
}

/* Iterate through all capture devices and
 * initialise them */
void capture_init_all(void)
{
	struct cap_s *cap;
	
	for(cap=captures; cap; cap=cap->next) {
		if ( cap->state != CAP_STATE_NULL ) continue;
		if ( (cap->priv=cap->capdev->init(cap->interface)) ) {
			cap->state=CAP_STATE_INIT;
		}else{
			cap->state=CAP_STATE_NULL;
		}
	}
}

/* Go through the first capture */
void capture_go(void)
{
	struct timeval tv, tv2;
	struct timeval end;
	
	if ( !captures || captures->state != CAP_STATE_INIT ) {
		cleanup(EXIT_ERR, "No captures loaded");
	}

	/* Ready to go */
	printf("capture: %s[%s]: started\n",
		captures->capdev->name,
		captures->interface ? captures->interface : "any");

	fflush(stdout);

	/* Do the capture */
	gettimeofday(&tv, NULL);
	captures->capdev->go(captures->priv, captures);
	gettimeofday(&tv2, NULL);
	captures->state=CAP_STATE_ENDED;

	/* All done */
	printf("capture: %s[%s]: stopped: %llu packets\n",
		captures->capdev->name,
		captures->interface ? captures->interface : "any",
		the_serial);

	/* Calculate elapsed times (badly) */
	end.tv_sec=(tv2.tv_sec-tv.tv_sec)-1;
	end.tv_usec=(1000000+tv2.tv_usec)-tv.tv_usec;
	while ( end.tv_usec > 1000000 ) {
		end.tv_usec-=1000000;
		end.tv_sec++;
	}

	printf("  BEGIN: %.lu.%.6lu\n", tv.tv_sec, tv.tv_usec);
	printf("    END: %.lu.%.6lu\n", tv2.tv_sec, tv2.tv_usec);
	printf("ELAPSED: %lu.%.6lu\n", end.tv_sec, end.tv_usec);
}
